"""Defines the configuration defaults and load functions used by `portray`"""
import ast
import os
import warnings
from typing import Any, Dict, List, Union, cast
from urllib import parse

import mkdocs.config as _mkdocs_config
import mkdocs.exceptions as _mkdocs_exceptions
from git import Repo
from toml import load as toml_load

import _ast
from portray.exceptions import NoProjectFound

PORTRAY_DEFAULTS = {
    "docs_dir": "docs",
    "output_dir": "site",
    "port": 8000,
    "host": "127.0.0.1",
    "labels": {"Cli": "CLI", "Api": "API", "Http": "HTTP", "Pypi": "PyPI"},
}

MKDOCS_DEFAULTS = {
    "site_name": os.path.basename(os.getcwd()),
    "config_file_path": os.getcwd(),
    "theme": {
        "name": "material",
        "palette": {"primary": "green", "accent": "lightgreen"},
        "custom_dir": os.path.join(os.path.dirname(__file__), "mkdocs_templates"),
    },
    "markdown_extensions": [
        "admonition",
        "codehilite",
        "extra",
        "pymdownx.details",
        "pymdownx.highlight",
    ],
}  # type: Dict[str, Any]

PDOCS_DEFAULTS = {
    "overwrite": True,
    "exclude_source": False,
    "template_dir": os.path.join(os.path.dirname(__file__), "pdocs_templates"),
}  # type: Dict[str, Union[str, str, bool, None, Dict, List]]


def project(directory: str, config_file: str, **overrides) -> dict:
    """Returns back the complete configuration - including all sub configuration components
       defined below that `portray` was able to determine for the project
    """
    if not (
        os.path.isfile(os.path.join(directory, config_file))
        or os.path.isfile(os.path.join(directory, "setup.py"))
        or "modules" in overrides
    ):
        raise NoProjectFound(directory)

    project_config = {**PORTRAY_DEFAULTS, "directory": directory}  # type: Dict[str, Any]
    if os.path.isfile(os.path.join(directory, "setup.py")):
        project_config.update(setup_py(os.path.join(directory, "setup.py")))

    project_config.update(toml(os.path.join(directory, config_file), **overrides))

    project_config.setdefault("modules", [os.path.basename(os.getcwd())])
    project_config.setdefault("pdocs", {}).setdefault("modules", project_config["modules"])

    project_config["mkdocs"] = mkdocs(directory, **project_config.get("mkdocs", {}))
    if "pdoc3" in project_config:
        warnings.warn("pdoc3 config usage is deprecated in favor of pdocs. "
                      "pdoc3 section will be ignored. ", DeprecationWarning)
    project_config["pdocs"] = pdocs(
        directory, **project_config.get("pdocs", {})
    )
    return project_config


def setup_py(location: str) -> dict:
    """Returns back any configuration info we are able to determine from a setup.py file"""
    setup_config = {}
    try:
        with open(location) as setup_py_file:
            for node in ast.walk(ast.parse(setup_py_file.read())):
                if (
                    type(node) == _ast.Call
                    and type(getattr(node, "func", None)) == _ast.Name
                    and node.func.id == "setup"  # type: ignore
                ):
                    for keyword in node.keywords:  # type: ignore
                        if keyword.arg == "packages":
                            setup_config["modules"] = ast.literal_eval(keyword.value)
                            break
                    break
    except Exception as error:
        warnings.warn(f"Error ({error}) occurred trying to parse setup.py file: {location}")

    return setup_config


def toml(location: str, **overrides) -> dict:
    """Returns back the configuration found within the projects
       [TOML](https://github.com/toml-lang/toml#toml) config (if there is one).

       Generally this is a `pyproject.toml` file at the root of the project
       with a `[tool.portray]` section defined.
    """
    try:
        toml_config = toml_load(location)
        tools = toml_config.get("tool", {})

        config = tools.get("portray", {})
        config.update(overrides)
        config["file"] = location

        if "modules" not in config:
            if "poetry" in tools and "name" in tools["poetry"]:
                config["modules"] = [tools["poetry"]["name"]]
            elif (
                "flit" in tools
                and "metadata" in tools["flit"]
                and "module" in tools["flit"]["metadata"]
            ):
                config["modules"] = [tools["flit"]["metadata"]["module"]]

        return config
    except Exception:
        warnings.warn(f"No {location} config file found")

    return {}


def repository(directory: str) -> dict:
    """Returns back any information that can be determined by introspecting the projects git repo
       (if there is one).
    """
    config = {}
    try:
        repo_url = Repo(directory).remotes.origin.url
        if "http" in repo_url:
            config["repo_url"] = repo_url
            config["repo_name"] = parse.urlsplit(repo_url).path.rstrip(".git").lstrip("/")
    except Exception:
        config = {}

    if not config:
        warnings.warn("Unable to identify `repo_name` and `repo_url` automatically")

    return config


def mkdocs(directory: str, **overrides) -> dict:
    """Returns back the configuration that will be used when running mkdocs"""
    mkdocs_config = {
        **MKDOCS_DEFAULTS,
        **repository(directory),
        **overrides,
    }  # type: Dict[str, Any]
    theme = mkdocs_config["theme"]
    if theme["name"].lower() == "material" and "custom_dir" not in theme:
        theme["custom_dir"] = MKDOCS_DEFAULTS["theme"]["custom_dir"]

    return mkdocs_config


def pdocs(directory: str, **overrides) -> dict:
    """Returns back the configuration that will be used when running pdocs"""
    defaults = {**PDOCS_DEFAULTS}
    defaults.update(overrides)
    return defaults



"""This module defines CLI interaction when using `portray`.

This is powered by [hug](https://github.com/hugapi/hug) which means unless necessary
it should maintain 1:1 compatibility with the programmatic API definition in the
[API module](/reference/portray/api)

- `portray as_html`: Renders the project as HTML into the `site` or other specified output directory
- `portray in_browser`: Runs a server with the rendered documentation pointing a browser to it
- `portray server`: Starts a local development server (by default at localhost:8000)
- `portray project_configuration`: Returns back the project configuration as determined by` portray`
"""
from pprint import pprint

import hug

from portray import api, logo

cli = hug.cli(api=hug.API(__name__, doc=logo.ascii_art))
cli(api.as_html)
cli.output(pprint)(api.project_configuration)
cli(api.server)
cli(api.in_browser)
cli(api.on_github_pages)



"""All portray specific exception classes should be defined here"""


class PortrayError(Exception):
    """Base class for all exceptions returned from portray"""

    pass


class NoProjectFound(PortrayError):
    """Thrown when portray is ran in a directory with no Python project"""

    def __init__(self, directory: str):
        super().__init__(
            self,
            f"No Python project found in the given directory: '{directory}'"
            + " See: https://timothycrosley.github.io/portray/TROUBLESHOOTING/#noprojectfound",
        )
        self.directory = directory


class DocumentationAlreadyExists(PortrayError):
    """Thrown when portray has been told to output documentation where it already exists"""

    def __init__(self, directory: str):
        super().__init__(
            self, f"Documentation already exists in '{directory}'. Use --overwrite to ignore"
        )
        self.directory = directory



"""This module defines the programmatic API that can be used to interact with `portray`
   to generate and view documentation.

   If you want to extend `portray` or use it directly from within Python - this is the place
   to start.
"""
import os
import webbrowser
from typing import Optional

import hug
import mkdocs.commands.gh_deploy

from portray import config, logo, render


def as_html(
    directory: str = "",
    config_file: str = "pyproject.toml",
    output_dir: str = "site",
    overwrite: bool = False,
    modules: list = None,
) -> None:
    """Produces HTML documentation for a Python project placing it into output_dir.

       - *directory*: The root folder of your project.
       - *config_file*: The [TOML](https://github.com/toml-lang/toml#toml)
         formatted config file you wish to use.
       - *output_dir*: The directory to place the generated HTML into.
       - *overwrite*: If set to `True` any existing documentation output will be removed
         before generating new documentation. Otherwise, if documentation exists in the
         specified `output_dir` the command will fail with a `DocumentationAlreadyExists`
         exception.
       - *modules*: One or more modules to render reference documentation for
    """
    directory = directory if directory else os.getcwd()
    render.documentation(
        project_configuration(directory, config_file, modules), overwrite=overwrite
    )
    print(logo.ascii_art)
    print(f"Documentation successfully generated into `{os.path.abspath(output_dir)}` !")


def in_browser(
    directory: str = "",
    config_file: str = "pyproject.toml",
    port: int = None,
    host: str = None,
    modules: list = None,
) -> None:
    """Opens your default webbrowser pointing to a locally started development webserver enabling
       you to browse documentation locally

       - *directory*: The root folder of your project.
       - *config_file*: The [TOML](https://github.com/toml-lang/toml#toml) formatted
         config file you wish to use.
       - *port*: The port to expose your documentation on (defaults to: `8000`)
       - *host*: The host to expose your documentation on (defaults to `"127.0.0.1"`)
       - *modules*: One or more modules to render reference documentation for
    """
    directory = directory if directory else os.getcwd()
    server(directory=directory, config_file=config_file, open_browser=True, modules=modules)


def server(
    directory: str = "",
    config_file: str = "pyproject.toml",
    open_browser: bool = False,
    port: int = None,
    host: str = None,
    modules: list = None,
) -> None:
    """Runs a development webserver enabling you to browse documentation locally.

       - *directory*: The root folder of your project.
       - *config_file*: The [TOML](https://github.com/toml-lang/toml#toml) formatted
         config file you wish to use.
       - *open_browser*: If true a browser will be opened pointing at the documentation server
       - *port*: The port to expose your documentation on (defaults to: `8000`)
       - *host*: The host to expose your documentation on (defaults to `"127.0.0.1"`)
       - *modules*: One or more modules to render reference documentation for
    """
    directory = directory if directory else os.getcwd()
    api = hug.API("Doc Server")

    project_config = project_configuration(directory, config_file, modules)
    with render.documentation_in_temp_folder(project_config) as doc_folder:

        @hug.static("/", api=api)
        def my_static_dirs():  # pragma: no cover
            return (doc_folder,)

        @hug.startup(api=api)
        def custom_startup(*args, **kwargs):  # pragma: no cover
            print(logo.ascii_art)
            if open_browser:
                webbrowser.open_new(f"http://{project_config['host']}:{project_config['port']}")

        api.http.serve(
            host=host or project_config["host"],
            port=port or project_config["port"],
            no_documentation=True,
            display_intro=False,
        )


def project_configuration(
    directory: str = "", config_file: str = "pyproject.toml", modules: list = None
) -> dict:
    """Returns the configuration associated with a project.

        - *directory*: The root folder of your project.
        - *config_file*: The [TOML](https://github.com/toml-lang/toml#toml) formatted
          config file you wish to use.
        - *modules*: One or more modules to include in the configuration for reference rendering
    """
    overrides = {}
    if modules:
        overrides["modules"] = modules
    directory = directory if directory else os.getcwd()
    return config.project(directory=directory, config_file=config_file, **overrides)


def on_github_pages(
    directory: str = "",
    config_file: str = "pyproject.toml",
    message: str = None,
    force: bool = False,
    ignore_version: bool = False,
    modules: list = None,
) -> None:
    """Regenerates and deploys the documentation to GitHub pages.

        - *directory*: The root folder of your project.
        - *config_file*: The [TOML](https://github.com/toml-lang/toml#toml) formatted
          config file you wish to use.
        - *message*: The commit message to use when uploading your documentation.
        - *force*: Force the push to the repository.
        - *ignore_version*: Ignore check that build is not being deployed with an old version.
        - *modules*: One or more modules to render reference documentation for
    """
    directory = directory if directory else os.getcwd()
    project_config = project_configuration(directory, config_file, modules)
    with render.documentation_in_temp_folder(project_config):
        conf = render._mkdocs_config(project_config["mkdocs"])
        conf.config_file_path = directory
        mkdocs.commands.gh_deploy.gh_deploy(
            conf, message=message, force=force, ignore_version=ignore_version
        )
        print(logo.ascii_art)
        print("Documentation successfully generated and pushed!")



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['portray']

package_data = \
{'': ['*'], 'portray': ['mkdocs_templates/partials/*', 'pdocs_templates/*']}

install_requires = \
['GitPython>=3.0,<4.0',
 'hug>=2.6,<3.0',
 'mkdocs-material>=4.4,<5.0',
 'mkdocs>=1.0,<2.0',
 'pdocs>=1.0,<2.0',
 'pymdown-extensions>=6.0,<7.0',
 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['portray = portray.cli:__hug__.cli']}

setup_kwargs = {
    'name': 'portray',
    'version': '1.2.1',
    'description': 'Your Project with Great Documentation',
    'long_description': "[![portray - Your Project with Great Documentation.](https://raw.github.com/timothycrosley/portray/master/art/logo.png)](https://timothycrosley.github.io/portray/)\n_________________\n\n[![PyPI version](https://badge.fury.io/py/portray.svg)](http://badge.fury.io/py/portray)\n[![Build Status](https://travis-ci.org/timothycrosley/portray.svg?branch=master)](https://travis-ci.org/timothycrosley/portray)\n[![codecov](https://codecov.io/gh/timothycrosley/portray/branch/master/graph/badge.svg)](https://codecov.io/gh/timothycrosley/portray)\n[![Join the chat at https://gitter.im/timothycrosley/portray](https://badges.gitter.im/timothycrosley/portray.svg)](https://gitter.im/timothycrosley/portray?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n[![License](https://img.shields.io/github/license/mashape/apistatus.svg)](https://pypi.python.org/pypi/portray/)\n[![Downloads](https://pepy.tech/badge/portray)](https://pepy.tech/project/portray)\n_________________\n\n[Read Latest Documentation](https://timothycrosley.github.io/portray/) - [Browse GitHub Code Repository](https://github.com/timothycrosley/portray/)\n_________________\n\n> The only thing worse than documentation never written, is documentation written but never discovered.\n\n**portray** is a Python3 command line tool and library that helps you create great documentation websites for your Python projects with as little effort as possible.\n\n[![Example Usage Gif](https://raw.githubusercontent.com/timothycrosley/portray/master/art/example.gif)](https://raw.githubusercontent.com/timothycrosley/portray/master/art/example.gif)\n\nKey Features:\n\n* **Zero-Config**: No configuration is necessary to use `portray`. Just run `portray` in the root of your Python project and it will find your documentation.\n* **Statically Generated**: `portray` websites are easy to host on GitHub pages and other similar services as they are outputted as standard static HTML websites.\n* **Markdown Aware**: `portray` will automatically include your projects `.md` files and render them into HTML. It will also find and render Markdown within `__doc__` strings.\n* **Fully Configurable**: While `portray` doesn't have to be configured, you still can fully configure it within the standard `pyproject.toml` file.\n* **Easily Programmable**: `portray` exposes a clean and simple [Python API](https://timothycrosley.github.io/portray/reference/portray/api/).\n* **Searchable**: Out of the box `portray` makes all of your documentation, even autogenerated code references, free-text searchable.\n* **Themeable**: `portray` is compatible with all existing [MkDocs Themes](https://github.com/mkdocs/mkdocs/wiki/MkDocs-Themes).\n\nUnder the hood, portray combines the [Markdown](https://commonmark.org/help/) documentation rendering capabilities provided by [MkDocs](https://www.mkdocs.org/)\nwith the automatic reference documentation generated by [pdocs](https://timothycrosley.github.io/pdocs/).\n\n## Quick Start\n\nThe following guides should get you up and running with a documentation website in no time.\n\n1. [Installation](https://timothycrosley.github.io/portray/docs/quick_start/1.-installation/) - TL;DR: Run `pip3 install portray` within your projects virtual environment.\n2. [Command Line Usage](https://timothycrosley.github.io/portray/docs/quick_start/2.-cli/) - TL;DR: Run `portray in_browser` to test and `portray on_github_pages` to deploy.\n3. [API Usage](https://timothycrosley.github.io/portray/docs/quick_start/3.-api/) - TL;DR: Everything available via the CLI is also easily available programmatically from within Python.\n4. [Configuration](https://timothycrosley.github.io/portray/docs/quick_start/4.-configuration/) - TL;DR: Put all configuration within a `[tool.portray]` section of your `pyproject.toml` file.\n\n## Why Create Portray?\n\nI create a lot of Python projects and I've always wanted a quick way to build documentation sites for them that included both manually written guides and autogenerated reference documentation.\nTools have improved over the years, but still nothing has given me this experience out of the box. Portray isn't a competitor for the 2 newer Python documentation projects (pdocs and MkDocs), rather it is melding of them\ntogether to give me the quick to create and comprehensive documentation website generation I've always wanted.\n\nThanks and I hope you too find `portray` useful!\n\n~Timothy Crosley\n",
    'author': 'Timothy Crosley',
    'author_email': 'timothy.crosley@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Defines how to render the current project and project_config using the
included documentation generation utilities.
"""
import os
import shutil
import tempfile
from argparse import Namespace
from contextlib import contextmanager
from glob import glob
from typing import Dict

import mako.exceptions
import mkdocs.config as mkdocs_config
import mkdocs.exceptions as _mkdocs_exceptions
from pdocs import as_markdown as pdocs_as_markdown
from mkdocs.commands.build import build as mkdocs_build

from portray.exceptions import DocumentationAlreadyExists

NO_HOME_PAGE = """
# Nothing here

`portray` uses README.md as your projects home page.
It appears you do not yet have a README.md file created.
"""


def documentation(config: dict, overwrite: bool = False) -> None:
    """Renders the entire project given the project config into the config's
    specified output directory.

    Behind the scenes:

    - A temporary directory is created and your code is copy and pasted there
    - pdoc is ran over your code with the output sent into the temporary directory
        as Markdown documents
    - MkDocs is ran over all of your projects Markdown documents including those
        generated py pdoc. MkDocs outputs an HTML representation to a new temporary
        directory.
    - The html temporary directory is copied into your specified output location
    - Both temporary directories are deleted.
    """
    if os.path.exists(config["output_dir"]):
        if overwrite:
            shutil.rmtree(config["output_dir"])
        else:
            raise DocumentationAlreadyExists(config["output_dir"])

    with documentation_in_temp_folder(config) as documentation_output:  # type: str
        shutil.copytree(documentation_output, config["output_dir"])


def pdocs(config: dict) -> None:
    """Render this project using the specified pdoc config passed into pdoc.

    This rendering is from code definition to Markdown so that
    it will be compatible with MkDocs.
    """
    pdocs_as_markdown(**config)


def mkdocs(config: dict):
    """Render the project's associated Markdown documentation using the specified
    MkDocs config passed into the MkDocs `build` command.

    This rendering is from `.md` Markdown documents into HTML
    """
    config_instance = _mkdocs_config(config)
    return mkdocs_build(config_instance)


@contextmanager
def documentation_in_temp_folder(config: dict):
    """Build documentation within a temp folder, returning that folder name before it is deleted."""
    with tempfile.TemporaryDirectory() as input_dir:
        input_dir = os.path.join(input_dir, "input")
        with tempfile.TemporaryDirectory() as temp_output_dir:
            shutil.copytree(config["directory"], input_dir)

            if "output_dir" not in config["pdocs"]:
                config["pdocs"]["output_dir"] = os.path.join(input_dir, "reference")
            pdocs(config["pdocs"])

            if "docs_dir" not in config["mkdocs"]:
                config["mkdocs"]["docs_dir"] = input_dir
            if "site_dir" not in config["mkdocs"]:
                config["mkdocs"]["site_dir"] = temp_output_dir
            if "nav" not in config["mkdocs"]:
                nav = config["mkdocs"]["nav"] = []

                root_docs = sorted(glob(os.path.join(input_dir, "*.md")))
                readme_doc = os.path.join(input_dir, "README.md")
                if readme_doc in root_docs:
                    root_docs.remove(readme_doc)
                else:
                    with open(readme_doc, "w") as readme_doc_file:
                        readme_doc_file.write(NO_HOME_PAGE)

                nav.append({"Home": "README.md"})

                nav.extend(_doc(doc, input_dir, config) for doc in root_docs)

                nav.extend(
                    _nested_docs(os.path.join(input_dir, config["docs_dir"]), input_dir, config)
                )

                reference_docs = _nested_docs(config["pdocs"]["output_dir"], input_dir, config)
                nav.append({"Reference": reference_docs})  # type: ignore

            mkdocs(config["mkdocs"])
            yield temp_output_dir


def _mkdocs_config(config: dict) -> mkdocs_config.Config:
    config_instance = mkdocs_config.Config(schema=mkdocs_config.DEFAULT_SCHEMA)
    config_instance.load_dict(config)

    errors, warnings = config_instance.validate()
    if errors:
        print(errors)
        raise _mkdocs_exceptions.ConfigurationError(
            f"Aborted with {len(errors)} Configuration Errors!"
        )
    elif config.get("strict", False) and warnings:
        print(warnings)
        raise _mkdocs_exceptions.ConfigurationError(
            f"Aborted with {len(warnings)} Configuration Warnings in 'strict' mode!"
        )

    config_instance.config_file_path = config["config_file_path"]
    return config_instance


def _nested_docs(directory: str, root_directory: str, config: dict) -> list:
    nav = [
        _doc(doc, root_directory, config) for doc in sorted(glob(os.path.join(directory, "*.md")))
    ]

    nested_dirs = sorted(glob(os.path.join(directory, "*/")))
    for nested_dir in nested_dirs:
        dir_nav = {
            _label(nested_dir[:-1], config): _nested_docs(nested_dir, root_directory, config)
        }
        nav.append(dir_nav)  # type: ignore

    return nav


def _label(path: str, config: Dict) -> str:
    label = os.path.basename(path)
    if "." in label:
        label = ".".join(label.split(".")[:-1])
    label = label.replace("-", " ").replace("_", " ").title()
    return config["labels"].get(label, label)


def _doc(path: str, root_path: str, config: dict) -> Dict[str, str]:
    path = os.path.relpath(path, root_path)
    return {_label(path, config): path}
